"""Text"""

from . import client
import warnings
from . import utils

from vbl_aquarium.models.urchin import TextModel
from vbl_aquarium.models.generic import IDData, IDListStringList, IDListColorList, IDListFloatList, IDListVector2List

## Text renderer

counter = 0

def clear():
    """Clear all custom meshes
    """
    client.sio.emit('Clear','text')

class Text:
  def __init__(self, text = "", color = [1, 1, 1], font_size = 12, position = [0,0]):
    global counter
    counter +=1

    self.data = TextModel(
      id = f't{counter}',
      text = text,
      color = utils.formatted_color(color),
      font_size=font_size,
      position=utils.formatted_vector2(position)
    )

    self._update()
    self.in_unity = True

  def _update(self):
    """Send serialized data to update this text object in Urchin
    """
    client.sio.emit('urchin-text-update', self.data.to_string())
  
  def delete(self):
    """Delete a text object
        
    Examples
    --------
    >>> t1.delete()
    """
    client.sio.emit('urchin-text-delete', IDData(id=self.data.id).to_string())
    self.in_unity = False

  def set_text(self, text):
    """Set the text in a set of text objects

    Parameters
    ----------
    text : string
      text to be displayed

    Examples
    --------
    >>> t1.set_text('test text')
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    
    self.data.text = utils.sanitize_string(text)
    self._update()

  def set_color(self,color):
    """Set the color of a set of text objects

    Parameters
    ----------
    color : color
        hex code or [R,G,B]
        
    Examples
    --------
    >>> t1.set_color('#FF0000')
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    
    self.data.color = utils.formatted_color(color)
    self._update()

  def set_font_size(self, font_size):
    """Set the font size of a set of text objects

    Parameters
    ----------
    text_sizes : int
        font sizes
        
    Examples
    --------
    >>> t1.set_size(12)
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    
    self.data.font_size = font_size
    self._update()

  def set_position(self,position):
    """Set the positions of a set of text objects in UI canvas space
    Bottom left corner is [-1,-1], top right [1,1]

    Text is anchored at the top left corner of its text box.

    Parameters
    ----------
    text_pos : list of two floats
        canvas positions relative to the center
        
    Examples
    --------
    >>> t1.set_position([400, 300])
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    
    self.data.position = utils.formatted_vector2(position)
    self._update()


def create(n):
  """Create n text objects with default parameters

  Parameters
  ----------
  n : int
      number of text objects
  """
  text_list = []
  for i in range(n):
    text_list.append(Text())
  return text_list

def set_texts(text_list, str_list):
  """Set the string value of multiple text objects

  Parameters
  ----------
  text_list : list of Text
      Text objects
  str_list : _type_
      _description_
  """
  str_list = utils.sanitize_list(str_list, len(text_list))

  for text, str, in zip(text_list, str_list):
    text.data.text = str

  data = IDListStringList(
    ids = [text.data.id for text in text_list],
    values= [string for string in str_list]
  )

  client.sio.emit('urchin-text-texts', data.to_string())

def set_positions(text_list, pos_list):
  """Set the positions of multiple text objects

  Positions are [0,1] relative to the edges of the screen

  Parameters
  ----------
  text_list : list of Text
      Text objects
  pos_list : list of float
      [0,0] top left [1,1] bottom right
  """
  pos_list = utils.sanitize_list(pos_list, len(text_list))
  
  for text, pos, in zip(text_list, pos_list):
    text.data.position = utils.formatted_vector2(pos)

  data = IDListVector2List(
    ids = [text.data.id for text in text_list],
    values = [text.data.position for text in text_list]
  )

  client.sio.emit('urchin-text-positions', data.to_string())

def set_font_sizes(text_list, font_size_list):
  """_summary_

  Parameters
  ----------
  text_list : list of Text
      Text objects
  font_size_list : _type_
      _description_
  """

  font_size_list = utils.sanitize_list(font_size_list, len(text_list))
  
  for text, font_size, in zip(text_list, font_size_list):
    text.data.font_size = font_size

  data = IDListFloatList(
    ids = [text.data.id for text in text_list],
    values= [text.data.font_size for text in text_list]
  )
  
  client.sio.emit('urchin-text-sizes', data.to_string())

def set_colors(text_list, color_list):
  """_summary_

  Parameters
  ----------
  text_list : list of Text
      Text objects
  color_list : _type_
      _description_
  """
  color_list = utils.sanitize_list(color_list, len(text_list))

  for text, color, in zip(text_list, color_list):
    text.data.color = utils.formatted_color(color)

  data = IDListColorList(
    ids = [text.data.id for text in text_list],
    values= [text.data.color for text in text_list]
  )
  
  client.sio.emit('urchin-text-colors', data.to_string())


from unittest import TestCase
from unittest.mock import Mock

import oursin as urchin


class TestClass(TestCase):
    """Test demo class functions"""

    # def setUp(self):
    #     """Setup class and mock callback function"""
    #     self.demo = DemoClass('Demo')
    #     self.mock = Mock()

    def test_sanitize_vec3(self):
        self.assertEqual(urchin.utils.sanitize_vector3([1,2,3]), [1,2,3])
        
        self.assertEqual(urchin.utils.sanitize_vector3((1,2,3)), [1,2,3])
        
        self.assertRaises(Exception, urchin.utils.sanitize_vector3, (1,2))


"""Camera"""

from . import client
from . import utils

import PIL
from PIL import Image

import numpy as np

import io
import json
import asyncio

from vbl_aquarium.models.urchin import CameraRotationModel, CameraModel
from vbl_aquarium.models.generic import FloatData, IDData, Vector2Data
			
receive_totalBytes = {}
receive_bytes = {}
receive_camera = {}

main = []

PIL.Image.MAX_IMAGE_PIXELS = 22500000

# Handle receiving camera images back as screenshots
def on_camera_img_meta(data_str):
	"""Handler for receiving metadata about incoming images

	Parameters
	----------
	data_str : string
		JSON with two fields {"name":"", "totalBytes":""}
	"""
	global receive_totalBytes

	data = json.loads(data_str)

	name = data["name"]
	totalBytes = data["totalBytes"]

	receive_totalBytes[name] = totalBytes
	receive_bytes[name] = bytearray()

def on_camera_img(data_str):
	"""Handler for receiving data about incoming images

	Parameters
	----------
	data_str : string
		JSON with two fields {"name":"", "bytes":""}
	"""
	global receive_totalBytes, receive_bytes, receive_camera

	data = json.loads(data_str)

	name = data["name"]
	byte_data = bytes(data["data"])

	receive_bytes[name] = receive_bytes[name] + byte_data
	
	if len(receive_bytes[name]) == receive_totalBytes[name]:
		print(f'(Camera receive) Camera {name} received an image')
		receive_camera[name].image_received = True

## Camera renderer
counter = 0

class Camera:
	def __init__(self, main = False):	
		global counter
		counter += 1	

		self.data = CameraModel(
			id = 'CameraMain' if main else f'Camera{counter}',
			controllable= True if main else False
		)
		
		self._update()

		self.in_unity = True
		self.image_received_event = asyncio.Event()
		self.loop = asyncio.get_event_loop()

	def _update(self):
		client.sio.emit('urchin-camera-update', self.data.to_string())

	def delete(self):
		"""Deletes camera
		
		Examples
		--------
		>>>c1.delete()
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
			
		client.sio.emit('urchin-camera-delete', IDData(id = self.data.id).to_string())
		self.in_unity = False

	def set_target_coordinate(self,camera_target_coordinate):
		"""Set the camera target coordinate in CCF space in um relative to CCF (0,0,0), without moving the camera. Coordinates can be negative or larger than the CCF space (11400,13200,8000)

		Parameters
		----------
		camera_target_coordinate : float list
			list of coordinates in ap, ml, dv in um

		Examples
		--------
		>>>c1.set_target_coordinate([500,1500,1000])
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
			
		self.data.target= utils.formatted_vector3(utils.sanitize_vector3(camera_target_coordinate))
		self._update()

	def set_rotation(self, rotation):
		"""Set the camera rotation (pitch, yaw, roll). The camera is locked to a target, so this rotation rotates around the target.

		Rotations are applied in order: roll, yaw, pitch. This can cause gimbal lock.

		Parameters
		----------
		rotation : float list OR string
			list of euler angles to set the camera rotation in (pitch, yaw, roll)
			OR
			string: "axial", "coronal", "sagittal", or "angled"

		Examples
		--------
		>>> c1.set_rotation([0,0,0])
		>>> c1.set_rotation("angled")
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		if rotation == 'axial':
			rotation = [0,0,0]
		elif rotation == 'sagittal':
			rotation = [0,90,-90]
		elif rotation == 'coronal':
			rotation = [-90,0,0]
		elif rotation == 'angled':
			rotation = [22.5,22.5,225]
		
		rotation = utils.sanitize_vector3(rotation)
		self.data.rotation= utils.formatted_vector3(rotation)
		
		self._update()

	def set_zoom(self,zoom):
		"""Set the camera zoom. 

		Parameters
		----------
		zoom : float	
			camera zoom parameter

		Examples
		--------
		>>> c1.set_zoom(1.0)
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		self.data.zoom=utils.sanitize_float(zoom)
		self._update()

	# def set_target_area(self, camera_target_area):
	# 	"""Set the camera rotation to look towards a target area

	# 	Note: some long/curved areas have mesh centers that aren't the 'logical' center. For these areas, calculate a center yourself and use set_camera_target.

	# 	Parameters
	# 	----------
	# 	camera_target_area : string
	# 		area ID or acronym, append "-lh" or "-rh" for one-sided meshes
	# 	Examples
	# 	--------
	# 	>>> c1.set_target_area("grey-l") 
	# 	"""
	# 	if self.in_unity == False:
	# 		raise Exception("Camera is not created. Please create camera before calling method.")
		
	# 	camera_target_area
	# 	self.target = camera_target_area
	# 	client.sio.emit('SetCameraTargetArea', {self.data.id: camera_target_area})

	def set_pan(self,pan_x, pan_y):
		"""Set camera pan coordinates

		Parameters
		----------
		pan_x : float
			x coordinate
		pan_y : float
			y coordinate
		
		Examples
		--------
		>>> c1.set_pan(3.0, 4.0)
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		self.data.pan= utils.formatted_vector2([pan_x, pan_y])
		self._update()

	def set_mode(self, mode):
		"""Set camera perspective mode

		Parameters
		----------
		mode : string
			"perspective" or "orthographic" (default)
		
		Examples
		--------
		>>> c1.set_mode('perspective')
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		if mode == 'orthographic':
				self.data.mode= CameraModel.CameraMode.orthographic
		else:
				self.data.mode= CameraModel.CameraMode.perspective
				
		self._update()

	def set_background_color(self, background_color):
		"""Set camera background color

		Parameters
		----------
		background_color : hex color string

		Examples
		--------
		>>> c1.set_background_color('#000000') # black background
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		self.data.background_color= utils.formatted_color(background_color)

		self._update()

	def set_controllable(self):
		"""Sets camera to controllable
		
		Examples
		--------
		>>> c1.set_controllable()
		"""
		if self.in_unity == False:
			raise Exception("Camera is not created. Please create camera before calling method.")
		
		self.data.controllable=True
		self._update()
		
	async def screenshot(self, size=[1024,768], filename = 'return'):
		"""Capture a screenshot, must be awaited

		Parameters
		----------
		size : list, optional
			Size of the screenshot, by default [1024,768]
		filename: string, optional
			Filename to save to, relative to local path
			
		Examples
		--------
		>>> await urchin.camera.main.screenshot()
		"""
		global receive_totalBytes, receive_bytes, receive_camera
		self.image_received_event.clear()
		self.image_received = False
		receive_camera[self.data.id] = self

		if size[0] > 15000 or size[1] > 15000:
			raise Exception('(urchin.camera) Screenshots can''t exceed 15000x15000')
			
		data = Vector2Data(
			id = self.data.id,
			value= utils.formatted_vector2(size)
		)
		
		client.sio.emit('urchin-camera-screenshot-request', data.to_string())

		while not self.image_received:
			await asyncio.sleep(0.1)
		# await self.image_received_event.wait()

		# image is here, reconstruct it
		img = Image.open(io.BytesIO(receive_bytes[self.data.id]))
		
		print(f'(Camera receive) {self.data.id} complete')
		del receive_totalBytes[self.data.id]
		del receive_bytes[self.data.id]
		del receive_camera[self.data.id]

		if not filename == 'return':
			img.save(filename)
		else:
			return img
		
	async def capture_video(self, file_name, callback = None,
						 start_rotation = None, end_rotation = None,
						 frame_rate = 30, duration = 5,
						 size = (1024,768),
						 test = False):
		"""Capture a video and save it to a file, must be awaited

		Can be used in two modes, either by specifying a callback(frame#) or a start/end_rotation

		Parameters
		----------
		file_name : string
			_description_
		callback : function, optional
			callback to execute *prior* to each frame, by default None
		start_rotation : [yaw, pitch, roll], optional
		end_rotation : [yaw, pitch, roll], optional
		frame_rate : int, optional
			frames per second, by default 30
		duration : int, optional
			seconds, by default 5
		size : tuple, optional
			screenshot size, by default (1024,768)

		Examples
		--------
		>>> await urchin.camera.main.capture_video('output.mp4', callback=my_callback_function)
		>>> await urchin.camera.main.capture_video('output.mp4', start_rotation=[22.5, 22.5, 225], end_rotation=[22.5, 22.5, 0])

		"""
		try:
			import cv2
		except:
			raise Exception('Please install cv2 by running `pip install opencv-python` in your terminal to use the Video features')
		
		fourcc = cv2.VideoWriter_fourcc(*'mp4v')
		out = cv2.VideoWriter(file_name, fourcc, frame_rate, size)

		n_frames = frame_rate * duration

		if start_rotation is not None:
			client.sio.emit('urchin-camera-lerp-set', CameraRotationModel(
				start_rotation=utils.formatted_vector3(start_rotation),
				end_rotation=utils.formatted_vector3(end_rotation)
			).to_string())

		for frame in range(n_frames):

			if callback is not None:
				callback(frame)

			if start_rotation is not None:
				perc = frame / n_frames

				client.sio.emit('urchin-camera-lerp', FloatData(
					id=self.data.id,
					value=perc
				).to_string())
			
			if not test:
				img = await self.screenshot([size[0], size[1]])
				image_array = np.array(img)
				image_array = cv2.cvtColor(image_array, cv2.COLOR_RGB2BGR)
				out.write(image_array)
		
		out.release()
		print(f'Video captured on {self.data.id} saved to {file_name}')


def set_light_rotation(angles):
	"""Override the rotation of the main camera light

	Parameters
	----------
	angles : vector3
		Euler angles of light
	"""
	angles = utils.sanitize_vector3(angles)
	print(angles)
	print(isinstance(angles,list))
	client.sio.emit('SetLightRotation', angles)

def set_light_camera(camera_name = None):
	"""Change the camera that the main light is linked to (the light will rotate the camera)

	Parameters
	----------
	camera_name : string, optional
		Name of camera to attach light to, by default None
	"""
	if (camera_name is None):
		client.sio.emit('ResetLightLink')
	else:
		client.sio.emit('SetLightLink', camera_name)

def set_brain_rotation(yaw):
	"""Set the brain's rotation, independent of the camera. This is useful when you want to animate
	a brain rotating for a video. You can set the camera angle you want, and then on each frame
	update the brain rotation, as needed.

	Parameters
	----------
	yaw : float
		Yaw angle for the brain, independent of the camera
	"""

	client.sio.emit('urchin-brain-yaw', FloatData(id='', value=yaw).to_string())

def setup():
	global main
	main = Camera(main = True)


"""Sanitizing inputs to send through API"""
import numpy as np
from enum import Enum

from vbl_aquarium.models.unity import *

### ENUMS
class Side(Enum):
    LEFT = -1
    FULL = 0
    RIGHT = 1
    ALL = 3

### SANITIZING FUNCTIONS

def sanitize_vector3(vector):
    """Guarantee that a vector is a vector 3, or raise an exception

    Parameters
    ----------
    input : any
        arbitrary input parameter

    Returns
    -------
    list
        vector3 as a list [x,y,z]

    Raises
    ------
    Exception
        Failed to coerce input to a length 3 list
    """
    try:
        vector_list = list(map(float, vector))
    except (TypeError, ValueError):
        raise ValueError("Input vector must be convertible to a list of three floats.")

    # Check if the length is exactly three
    if len(vector_list) != 3:
        raise ValueError("Input vector must have exactly three elements.")

    return vector_list


def sanitize_color(color):
    """
    Convert input color to a list of r/g/b/a values in the range 0->1.

    Parameters
    ----------
    color : str or list
        Hex code or list of floats representing color values.

    Returns
    -------
    list
        List of r/g/b/a values in the range 0->1.
    """
    if isinstance(color, list):
        if len(color) == 3 or len(color) == 4:
            try:
                if max(color) > 1:
                    return [x / 255 for x in color]
                else:
                    return color
            except (TypeError, ValueError):
                raise ValueError("Input list must contain three or four floats.")

    elif isinstance(color, str):
        if color.startswith('#'):
            return hex_to_rgb(color)

    else:
        raise TypeError("Input type not recognized.")

def sanitize_float(value):
    if isinstance(value, float):
        return value
    else:
        try:
            return float(value)
        except:
            raise Exception("Value could not be coerced to a float.")

def sanitize_material(material):
    if isinstance(material, str):
        return(material)
    else:
        raise Exception("Material is not properly passed in as a string. Please pass in material as a string.")

def sanitize_list(input, length=0):
    """Guarantee that a list is of at least size length, or try to broadcast to that size

    Parameters
    ----------
    input : list
    length : int, optional
        length to broadcast to, by default 0

    Returns
    -------
    list
    """
    if length > 0 and not isinstance(input, list):
        input = [input] * length

    if not isinstance(input, list):
        raise Exception("List parameter needs to be a list.")

    return input
    
def sanitize_string(string):
    if isinstance(string, str):
        return(string)
    else:
        raise Exception("Input is not properly passed in as a string. Please pass in input as a string.")
    
def sanitize_side(acronym, sided):
    if sided == "full":
        return acronym
    elif sided == "left":
        return f'{acronym}-lh'
    elif sided == "right":
        return f'{acronym}-rh'
    else:
        raise Exception(f'Sided enum {sided} not properly defined, should be full/left/right')
    
def rgb_to_hex(rgb):
    return '#%02x%02x%02x' % rgb
    
def rgba_to_hex(rgba):
    return '#%02x%02x%02x%02x' % rgba

def hex_to_rgb(hex_code):
    # Remove '#' if present
    hex_code = hex_code.lstrip('#')
    # Convert hexadecimal to RGB
    r = int(hex_code[0:2], 16)
    g = int(hex_code[2:4], 16)
    b = int(hex_code[4:6], 16)
    return (r/255, g/255, b/255)

def list_of_list2vector3(list_of_list):
    """Convert a list of lists to a list of Vector3 objects

    Parameters
    ----------
    list_of_list : list of length 3 lists
        _description_
    """
    return [formatted_vector3(data) for data in list_of_list]

def formatted_vector3(list_of_float):
    """Convert a list of floats to a Vector3

    Parameters
    ----------
    list_of_float : list
    """

    list_of_float = sanitize_vector3(list_of_float)

    return Vector3(
        x = list_of_float[0],
        y = list_of_float[1],
        z = list_of_float[2]
    )

def formatted_vector2(list_of_float):
  """Convert a list of floats to a Vector2

  Parameters
  ----------
  list_of_float : list
  """
  return Vector2(
      x = list_of_float[0],
      y = list_of_float[1]
  )

def formatted_color(color):
    """Converts a color, either a hex or list of floats, to a Color object

    Parameters
    ----------
    color : list/str
        Length 3 for RGB, 4 for RGBA, or a hex color string
    """

    color = sanitize_color(color)

    if len(color) == 3:
        return Color(
            r = color[0],
            g = color[1],
            b = color[2]
        )
    elif len(color) == 4:
        return Color(
            r = color[0],
            g = color[1],
            b = color[2],
            a = color[3]
        )
    else:
        raise Exception('Colors should be length 3 or 4')


from .. import client
from .. import utils
from pathlib import Path

import json
from vbl_aquarium.models.urchin import AtlasModel, StructureModel, ColormapModel
from vbl_aquarium.models.generic import *

class CustomAtlas:
    def __init__(self, atlas_name, atlas_dimensions, atlas_resolution):
        self.atlas_name = atlas_name

        data = {}
        data['name'] = atlas_name
        data['dimensions'] = utils.sanitize_vector3(atlas_dimensions)
        data['resolution'] = utils.sanitize_vector3(atlas_resolution)

        print(data)

        client.sio.emit('CustomAtlas', json.dumps(data))

class Atlas:
    def __init__(self, atlas_name):
        # load the ontology structure file
        self.loaded = False
        self.data = AtlasModel(
            name = atlas_name,
            areas = [],
            colormap=ColormapModel()
        )

        current_script_directory = Path(__file__).resolve().parent
        
        data_file_path = f'{current_script_directory}/data/{self.data.name}.structures.json'
        with open(data_file_path,'r') as f:
            temp = json.load(f)

        for i, structure_data in enumerate(temp):
            # {"acronym": "root", "id": 997, "name": "root", "structure_id_path": [997], "rgb_triplet": [255, 255, 255]}
            rgb_normalized = [x/255 for x in structure_data['rgb_triplet']]
            area_data = StructureModel(
                name = structure_data['name'],
                acronym = structure_data['acronym'],
                atlas_id = structure_data['id'],
                color= utils.formatted_color(rgb_normalized)
            )
            self.data.areas.append(area_data)

            area = Structure(
                data=area_data,
                index = i,
                update_callback=self._update
            )

            setattr(self, structure_data['acronym'], area)

    def _update(self):
        """Internal helper function, push data to Unity and update all values
        """
        client.sio.emit('urchin-atlas-update', self.data.to_string())

    def load(self):
        """Load this atlas
        """
        if self.loaded:
            print("(Warning) Atlas was already loaded, the renderer can have issues if you try to load an atlas twice.")
        
        self.loaded = True
        client.sio.emit('urchin-atlas-load', self.data.to_string())

    def clear(self):
        """Clear all visible areas
        """

        # update all areas to be not visible
        for area in self.data.areas:
            area.visible = False

        client.sio.emit('Clear', 'area')

    def load_defaults(self):
        """Load the left and right areas

        Note that this function is not stateful, if you save the scene it will not be reloaded.
        """
        client.sio.emit('urchin-atlas-defaults', "")

    def set_reference_coord(self, reference_coord):
        """Set the reference coordinate for the atlas (Bregma by default)

        Parameters
        ----------
        reference_coord : list of float
        """
        self.data.reference_coord = utils.formatted_vector3(utils.sanitize_vector3(reference_coord))
        self._update

    def get_areas(self, area_list):
        """Get the area objects given a list of area acronyms

        Parameters
        ----------
        area_list : list of string
            List of acronyms to get objects for

        Returns
        -------
        list of Structure

        Examples
        --------
        >>> area_list = urchin.get_areas(["root", "VISp"])
        """
        areas = []
        for name in area_list:
            try:
                area = getattr(self, name)
                areas.append(area)
            except:
                print(f'(Warning): Area {name} couldn''t be found in this atlas!')
        return areas

    def set_visibilities(self, area_list, area_visibility, side = utils.Side.FULL, push = True):
        """Set visibility of multiple areas at once

        Parameters
        ----------
        area_list : list of Structure
        area_visibility : list of bool
        sided : string, optional
            Brain area side to load, default = FULL

        Examples
        --------
        >>> urchin.ccf25.set_visibilities(urchin.ccf25.get_areas(["root", "VISp"]), [True, False])
        >>> urchin.ccf25.set_visibilities(urchin.ccf25.get_areas(["root", "VISp"]), True)
        """
        area_visibility = utils.sanitize_list(area_visibility, len(area_list))

        for i, area in enumerate(area_list):
            self.data.areas[area.index].visible = area_visibility[i]
            self.data.areas[area.index].side = side.value

        if push:
            self._update()

    def set_colors(self, area_list, area_colors, push = True):
        """Set color of multiple areas at once.

        Parameters
        ----------
        area_list : list of Structure
        area_colors : list of colors (hex string or RGB triplet)

        Examples
        --------
        >>> urchin.ccf25.set_visibilities(urchin.ccf25.get_areas(["root", "VISp"]), ['#ff0000', '#00ff00'])
        >>> urchin.ccf25.set_visibilities(urchin.ccf25.get_areas(["root", "VISp"]), [255, 255, 255])
        """
        area_colors = utils.sanitize_list(area_colors, len(area_list))
        
        for i, area in enumerate(area_list):
            self.data.areas[area.index].color = utils.formatted_color(area_colors[i])

        if push:
            self._update()
        
    def set_colormap(self, colormap_name, min = 0, max = 1):
        """Set colormap used for mapping area *intensity* values to colors


        Options are
        - cool (default, teal 0 -> magenta 1)
        - grey (black 0 -> white 1)
        - grey-green (grey 0, light 1/255 -> dark 1)
        - grey-purple (grey 0, light 1/255 -> dark 1)
        - grey-red (grey 0, light 1/255 -> dark 1)
        - grey-rainbow (grey 0, rainbow colors from 1/255 -> 1)

        Parameters
        ----------
        colormap_name : string
            colormap name
        """
        self.data.colormap = ColormapModel(
            name = colormap_name,
            min = min,
            max = max
        )
        self._update()

    def set_color_intensity(self, area_list, area_intensities, push = True):
        """Set intensity values, colors will be set according to the active colormap

        Parameters
        ----------
        area_intensities : list of float
            0->1

        Examples
        --------
        >>> urchin.ccf25.set_color_intensity(urchin.ccf25.get_areas(["root", "VISp"]), [0, 1])
        """
        area_intensities = utils.sanitize_list(area_intensities, len(area_list))

        for i, area in enumerate(area_list):
            self.data.areas[area.index].color_intensity = area_intensities[i]

        if push:
            self._update()

    def set_alphas(self, area_list, area_alphas, push = True):
        """Set alpha values, without changing colors

        Parameters
        ----------
        area_list : list of Structure
        area_alphas : list of float
        """
        area_alphas = utils.sanitize_list(area_alphas, len(area_list))

        for i, area in enumerate(area_list):
            self.data.areas[area.index].color.a = area_alphas[i]

        if push:
            self._update()

    def set_materials(self, area_list, area_materials, push = True):
        """Set material of multiple areas at once.

        Material options are
        - 'opaque-lit' or 'default'
        - 'opaque-unlit'
        - 'transparent-lit'
        - 'transparent-unlit'
        
        Parameters
        ----------
        area_colors : dict {string: string}
            Keys are area IDs or acronyms, Values are hex colors

        Examples
        --------
        >>> urchin.ccf25.set_materials(urchin.ccf25.get_areas(["root", "VISp"]), ['transparent-lit', 'opaque-lit'])
        >>> urchin.ccf25.set_materials(urchin.ccf25.get_areas(["root", "VISp"]), ['transparent-lit])
        """
        area_materials = utils.sanitize_list(area_materials, len(area_list))

        for i, area in enumerate(area_list):
            self.data.areas[area.index].material = area_materials[i]

        if push:
            self._update()

class Structure:
    """Structure attributes can be accessed as

    >>> structure.name
    >>> structure.acronym
    >>> structure.id
    >>> structure.rgb_triplet
    >>> structure.path
    """
    def __init__(self, data, index, update_callback):
        self.data = data
        self.index = index
        self.update_callback = update_callback

    def set_visibility(self, visibility, side = utils.Side.FULL, push = True):
        """Set area visibility

        Parameters
        ----------
        visibility : bool
        side : utils.Side, optional
            .FULL, .LEFT, or .RIGHT

        Examples
        --------
        >>> urchin.ccf25.root.set_visibility(True)
        >>> urchin.ccf25.root.set_visibility(True, urchin.utils.Side.LEFT)
        """
        self.data.visible = visibility
        self.data.side = side.value

        if push:
            self.update_callback()

    def set_color(self, color, push = True):
        """Set area color.

        Parameters
        ----------
        color : hex string

        Examples
        --------
        >>> urchin.ccf25.root.set_color('#ff0000')
        >>> urchin.ccf25.root.set_color([255, 0, 0], "left")
        """
        self.data.color = utils.formatted_color(utils.sanitize_color(color))

        if push:
            self.update_callback()

    def set_alpha(self, alpha, push = True):
        """Set area transparency.

        Parameters
        ----------
        alpha: float

        Examples
        --------
        >>> urchin.ccf25.root.set_alpha(0.5)
        """
        self.data.color.a = utils.sanitize_float(alpha)

        if push:
            self.update_callback()


    def set_intensity(self, intensity, push = True):
        """Set color based on the intensity value through the active colormap.

        Parameters
        ----------
        intensity : float
            0->1

        Examples
        --------
        >>> urn.set_intensity(0.5)
        """
        self.data.color_intensity = utils.sanitize_float(intensity)

        if push:
            self.update_callback()

    def set_material(self, material, push = True):
        """Set material.

        Material options are
        - 'opaque-lit' or 'default'
        - 'opaque-unlit'
        - 'transparent-lit'
        - 'transparent-unlit'

        Parameters
        ----------
        material: string

        Examples
        ----------
        >>> urchin.ccf25.root.set_material('transparent-lit')
        """
        self.data.material = utils.sanitize_string(material)

        if push:
            self.update_callback()

    # def set_data(area_data):
    #     """Set the data array for each CCF area model

    #     Data arrays work the same as the set_area_intensity() function but are controlled by the area_index value, which can be set in the renderer or through the API.

    #     Parameters
    #     ----------
    #     area_data : dict {string: float list}
    #         keys area IDs or acronyms, values are a list of floats
    #     """
    #     client.sio.emit('SetAreaData', area_data)

    # def set_data_index(area_index):
    #     """Set the data index for the CCF area models

    #     Parameters
    #     ----------
    #     area_index : int
    #         data index
    #     """
    #     client.sio.emit('SetAreaIndex', area_index)


"""Interactive components within notebooks"""
from . import particles
from . import meshes
from vbl_aquarium.models.generic import IDListFloatList

global_binned_spikes = None
global_prepped_data = None
neurons = None
neuron_colors = None
stim_id = None

n_fig = None
n_ax = None
n_vline = None

s_fig = None
s_ax = None
s_vline = None
# def slider_widget(function_call, slider_parameters):
#     """Creates a slider in the notebook, displays results of the input function
    
    
#     Parameters
#     ----------
#     function_call: function established earlier in code
#     slider_parameters: list of start, stop, and increment for slider to follow
    
#     Examples
#     ---------
#     >>> urchin.ui.slider_widget(update_sizes_from_firing)
#     """
#     try:
#         import ipywidgets as widgets
#     except ImportError:
#         raise ImportError("Widgets package is not installed. Please pip install ipywidgets to use this function.")
#     widgets.interact(function_call, t=(slider_parameters[0],slider_parameters[1],slider_parameters[2]))

def spikes_bin_data(spike_times_raw_data, spike_clusters_data, bin_size=0.02):
    """Bins spike clusters into an array
    
    Parameters
    ----------
    spike_times_raw_data: np array
        raw data of spiking times, in samples
    spike_clusters_data: np array
        list of spike clusters data
    bin_size: float
        bin size in seconds, default value is 20ms
    
    Returns
    -------
    array
        spiking data binned in given bin size
    
    Examples
    --------
    >>> urchin.ui.spikes_bin_data(spike_times_samp, spike_clusters)
    """
    try:
        import numpy as np
    except ImportError:
        raise ImportError("Numpy package is not installed. Please pip install numpy to use this function.")
    # bin the spike times and clusters
    spike_times_raw_data = np.squeeze(spike_times_raw_data)
    spike_clusters_data = np.squeeze(spike_clusters_data)
    spike_times_sec = spike_times_raw_data / 3e4 # convert from 30khz samples to seconds
    # set up bin edges - 20 ms here
    bins_seconds = np.arange(np.min(spike_times_sec), np.max(spike_times_sec), bin_size)
    # make list of lists for spike times specific to each cluster
    spikes = [spike_times_sec[spike_clusters_data == cluster] for cluster in np.unique(spike_clusters_data)]
    # bin
    binned_spikes = []
    for cluster in spikes:
        counts, _ = np.histogram(cluster, bins_seconds)  
        binned_spikes.append(counts)
    binned_spikes = np.array(binned_spikes) # should be [#neurons, #bins]
    global global_binned_spikes 
    global_binned_spikes = binned_spikes
    return binned_spikes

def spikes_binned_event_average(event_start, event_ids, binned_spikes = None, bin_size_sec=0.02, window_start_sec = 0.1, window_end_sec = 0.5):
    """Prepares intermediate data table and averages binned spikes over a given time window
    
    Parameters
    ----------
    binned_spikes: array
        binned spiking data
    event_start: array
        start times of events in seconds
    event_ids: array
        ids of events
    bin_size_sec: float
        bin size in seconds, default value is 20ms
    window_start_sec: float
        start of window in seconds, default value is 0.1
    window_end_sec: float
        end of window in seconds, default value is 0.5
    
    Returns
    -------
    array
        binned spikes averaged over given time window
    
    Examples
    --------
    >>> urchin.ui.spikes_binned_event_average(event_start, event_ids, binned_spikes)
    """
    try:
        import numpy as np
    except ImportError:
        raise ImportError("Numpy package is not installed. Please pip install numpy to use this function.")
    global global_binned_spikes

    if binned_spikes is None:
        binned_spikes = global_binned_spikes
    
    bintime_prev = int(window_start_sec * 50)
    bintime_post = int(window_end_sec * 50 + 1)
    windowsize = bintime_prev + bintime_post
    bin_size = bin_size_sec * 1000

    # To bin: divide by 20, floor
    stim_binned = np.floor(event_start * 1000 / bin_size).astype(int)
    stim_binned = np.transpose(stim_binned)


    u_stim_ids = np.unique(event_ids)

    # Initialize final_avg matrix
    final_avg = np.empty((binned_spikes.shape[0], len(u_stim_ids), windowsize))

    for neuron_id in range(binned_spikes.shape[0]):

        for stim_id in u_stim_ids:
            stim_indices = np.where(event_ids[0] == stim_id)[0]

            neuron_stim_data = np.empty((len(stim_indices), windowsize))
            
            for i, stim_idx in enumerate(stim_indices):
                bin_id = int(stim_binned[0][stim_idx])
                selected_columns = binned_spikes[neuron_id, bin_id - bintime_prev: bin_id + bintime_post]
                neuron_stim_data[i,:] = selected_columns

            bin_average = np.mean(neuron_stim_data, axis=0)/bin_size_sec
            final_avg[neuron_id, int(stim_id) - 1, :] = bin_average

    final_avg = np.delete(final_avg, 88, axis=0) #DROPPING THE NEURON WITH THE HIGH BASELINE

    global global_prepped_data
    global_prepped_data = final_avg
    return final_avg

def slope_viz_stimuli_per_neuron(change, prepped_data = None, t=-100):
    """Visualizes and creates interactive plot for the average of each stimulus per neuron
    
    Parameters
    ----------
    prepped_data: 3D array
        prepped data of averages of binned spikes and events in the format of [neuron_id, stimulus_id, time]
    t: int
        time in milliseconds of where to initially place the vertical line
    neuron_id: int
        id of neuron
    
    Examples
    --------
    >>> urchin.ui.slope_viz_stimuli_per_neuron(t=-100, neuron_id = 0)
    """
    try:
        import numpy as np
    except ImportError:
        raise ImportError("Numpy package is not installed. Please pip install numpy to use this function.")
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError("Matplotlib package is not installed. Please pip install matplotlib to use this function.")
    
    global global_prepped_data
    global n_fig, n_ax, n_vline
    if prepped_data is None:
        prepped_data = global_prepped_data

    if isinstance(change, int):
        neuron_id = change
    else:
        neuron_id = change.new

    # Plotting data:
    n_ax.clear()
    for i in range(0,prepped_data.shape[1]):
        y = prepped_data[neuron_id][i]
        x = np.arange(-100, 520, step=20)
        n_ax.plot(x,y, color='dimgray')

    # Labels:
    n_ax.set_xlabel('Time from stimulus onset')
    n_ax.set_ylabel('Number of Spikes Per Second')
    n_ax.set_title(f'Neuron {neuron_id} Spiking Activity with Respect to Each Stimulus')

    #Accessories:
    n_ax.axvspan(0, 300, color='gray', alpha=0.3)
    n_vline = n_ax.axvline(t, color='red', linestyle='--',)
    # Set y-axis limits
     # Calculate y-axis limits
    max_y = max([max(prepped_data[neuron_id][i]) for i in range(10)])  # Maximum y-value across all lines
    if max_y < 10:
        max_y = 10  # Set ymax to 10 if the default max is lower than 10
    n_ax.set_ylim(0, max_y)
   
    # plt.legend()
    # plt.show()
    
def update_neuron_sizing(stim_id, t, prepped_data = None):
        global neurons
        global global_prepped_data
    
        if prepped_data is None:
            prepped_data = global_prepped_data

        t_id = round((t+100)/20)
            
        size_list = []
        for i in range(prepped_data.shape[0]):
             size = round(prepped_data[i][stim_id][t_id]/200,4)
             size_list.append([size, size, size])


        meshes.set_scales(neurons, size_list)


        # particle_size_list = IDListFloatList(
        #     ids = [],
        #     values= []
        # )

        # for i in range(prepped_data.shape[0]):
        #     neuron = f'n{str(i+1)}'
        #     particle_size_list.ids.append(neuron)
        #     particle_size_list.values.append(round(prepped_data[i][stim_id][t_id]/200,4))
        
        
        # particles._set_sizes(particle_size_list)

def slope_viz_neurons_per_stimuli(change, prepped_data = None, n_color = None, t=-100):
    """Visualizes and creates interactive plot for the average of every neuron per stimulus
    
    Parameters
    ----------
    prepped_data: 3D array
        prepped data of averages of binned spikes and events in the format of [neuron_id, stimulus_id, time]
    t: int
        time in milliseconds of where to initially place the vertical line
    stim_id: int
        id of neuron
    
    Examples
    --------
    >>> urchin.ui.slope_viz_stimuli_per_neuron(t=-100, stim_id = 0)
    """
    try:
        import numpy as np
    except ImportError:
        raise ImportError("Numpy package is not installed. Please pip install numpy to use this function.")
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError("Matplotlib package is not installed. Please pip install matplotlib to use this function.")
     
    global global_prepped_data
    if prepped_data is None:
        prepped_data = global_prepped_data

    global neuron_colors
    if n_color is None:
        n_color = neuron_colors

    global s_fig, s_ax, s_vline, stim_id

    if isinstance(change, int):
        stim_id = change
    else:
        stim_id = change.new

     # Plotting data:
    s_ax.clear()
    for i in range(0,prepped_data.shape[0]):
        y = prepped_data[i][stim_id]
        x = np.arange(-100, 520, step=20)
        s_ax.plot(x,y, color = n_color[i])
    
    # Labels:
    s_ax.set_xlabel(f'Time from Stimulus {stim_id} display (20 ms bins)')
    s_ax.set_ylabel('Number of Spikes Per Second')
    s_ax.set_title(f'Neuron Spiking Activity with Respect to Stimulus ID {stim_id}')

    # Accessories:
    s_ax.axvspan(0, 300, color='gray', alpha=0.3)
    s_vline = s_ax.axvline(t, color='red', linestyle='--',)

    # plt.show()

    # update_neuron_sizing(stim_id, t)

def update_nline(position):
    global n_vline, n_fig
    position = position.new
    n_vline.set_xdata([position, position])  # Update x value of the vertical line
    n_fig.canvas.draw_idle()

def update_sline(t):
    global s_vline, s_fig, stim_id
    t = t.new
    s_vline.set_xdata([t, t])  # Update x value of the vertical line
    s_fig.canvas.draw_idle()
    global stim_id
    update_neuron_sizing(stim_id, t)

def plot_appropriate_interactive_graph(prepped_data = None, view = "stim", window_start_sec = 0.1, window_end_sec = 0.5):
    """Plots appropriate interactive graph based on view
    
    Parameters
    ----------
    prepped_data: 3D array
        prepped data of averages of binned spikes and events in the format of [neuron_id, stimulus_id, time]
    view: str
        view type, either "stim" or "neuron"
    window_start_sec: float
        start of window in seconds, default value is 0.1
    window_end_sec: float
        end of window in seconds, default value is 0.5
    
    Examples
    --------
    >>> urchin.ui.plot_appropriate_interactie_graph(prepped_data, view = "stim")
    """
    try:
        import ipywidgets as widgets
    except ImportError:
        raise ImportError("Widgets package is not installed. Please pip install ipywidgets to use this function.")
    
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError("Matplotlib package is not installed. Please pip install matplotlib to use this function.")
    
    from IPython.display import display
        
    global global_prepped_data
    if prepped_data is None:
        prepped_data = global_prepped_data
    
    
    
    if view == "stim":
        global s_fig, s_ax
        s_fig, s_ax = plt.subplots()

        time_slider = widgets.IntSlider(value=-1e3 * window_start_sec, min=-1e3 * window_start_sec, max=5e3 * window_start_sec, step=5, description='Time')
        time_slider.layout.width = '6.53in'
        time_slider.layout.margin = '0 -4px'

        stimuli_dropdown = widgets.Dropdown(
            options= range(0,prepped_data.shape[1]),
            value=0,
            description='Stimulus ID:',
        )
        stimuli_dropdown.layout.margin = "20px 20px"
        # output = widgets.interactive_output(slope_viz_neurons_per_stimuli, {'t': time_slider, 'stim_id': stimuli_dropdown})
        # # Display the widgets and the output
        # display(widgets.VBox([stimuli_dropdown,time_slider]))
        # display(output)

        ui = widgets.VBox([stimuli_dropdown,time_slider])
        slope_viz_neurons_per_stimuli(stimuli_dropdown.value)
        time_slider.observe(update_sline, names = "value")
        stimuli_dropdown.observe(slope_viz_neurons_per_stimuli, names = "value")
        display(ui)
    
    elif view == "neuron":
        global n_fig, n_ax
        n_fig, n_ax = plt.subplots()

        time_slider = widgets.IntSlider(value=-1e3 * window_start_sec, min=-1e3 * window_start_sec, max=5e3 * window_start_sec, step=5, description='Time')
        time_slider.layout.width = '6.53in'
        time_slider.layout.margin = '0 -4px'

        neuron_dropdown = widgets.Dropdown(
            options= range(0,prepped_data.shape[0]),
            value=354,
            description='Neuron ID:',
        )
        neuron_dropdown.layout.margin = "20px 20px"

        # # Link the function with the interact function
        # output = widgets.interactive_output(slope_viz_stimuli_per_neuron, {'t': time_slider, 'neuron_id': neuron_dropdown})

        # # Display the widgets and the output
        # display(widgets.VBox([neuron_dropdown,time_slider]))
        # display(output)

        ui = widgets.VBox([neuron_dropdown, time_slider])
        slope_viz_stimuli_per_neuron(neuron_dropdown.value)
        time_slider.observe(update_nline, names='value')
        neuron_dropdown.observe(slope_viz_stimuli_per_neuron,names='value')
        display(ui)
    

def plot_event_average_interaction(spiking_times, spiking_clusters, event_start, event_ids, view = "stim"):
    """Wrapper function that takes in raw data, and goes through entire process to return plots
    
    Parameters
    ----------
    spike_times_raw_data: np array
        raw data of spiking times, in samples
    spike_clusters_data: np array
        list of spike clusters data
    event_start: array
        start times of events in seconds
    event_ids: array
        ids of events
    view: str
        view type, either "stim" or "neuron"
    Examples
    --------
    >>> urchin.ui.plot_event_average_interaction(spike_times,spike_clusters,event_start, event_ids, view = "neuron")
    """
    binned_spikes = spikes_bin_data(spiking_times, spiking_clusters)
    prepped_data = spikes_binned_event_average(event_start, event_ids, binned_spikes)
    plot_appropriate_interactive_graph(prepped_data, view)


"""Client for communicating with the echo server"""
import socketio
import uuid
import asyncio

from . import camera
from . import volumes
from . import meshes
from . import dock

from vbl_aquarium.models.logging import *

class bcolors:
    WARNING = '\033[93m'
    FAIL = '\033[91m'

ID = str(uuid.uuid1())[:8]

sio = socketio.Client()
@sio.event
def connect():
	print("(URN) connected to server")
	change_id(ID)

@sio.event
def disconnect():
    print("(URN) disconnected from server")

@sio.on('log')
def message(data):
	print(data)
	out = Log.model_validate_json(data)
	print('(Renderer) ' + out.msg)

@sio.on('log-warning')
def message(data):
	out = LogWarning.model_validate_json(data)
	print('(Renderer) ' + bcolors.WARNING + out.msg)

@sio.on('log-error')
def message(data):
	out = LogError.model_validate_json(data)
	print('(Renderer) ' + bcolors.FAIL + out.msg)

###### CALLBACKS #######

@sio.on('CameraImgMeta')
def receive_camera_img_meta(data):
	camera.on_camera_img_meta(data)
	
@sio.on('CameraImg')
def receive_camera_img(data):
	camera.on_camera_img(data)

@sio.on('VolumeClick')
def receive_volume_click(data):
	volumes._volume_click(data)

@sio.on('NeuronCallback')
def receive_neuron_callback(data):
	meshes._neuron_callback(data)

@sio.on('urchin-dock-callback')
def receive_dock_callback(data):
	dock._save_callback(data)
	
# Helper functions
def connected():
	return sio.connected

def close():
	"""Disconnect from the echo server
	"""
	sio.disconnect()

def change_id(newID):
	"""Change the ID used to connect to the echo server

	Parameters
	----------
	newID : string
		New ID to connect with
	"""
	sio.emit('ID',[newID,"send"])
	print(f'Login sent with ID: {newID}, copy this ID into the renderer to connect.')



"""Particles"""

from . import client
import warnings
from . import utils

from vbl_aquarium.models.unity import Vector3, Color
from vbl_aquarium.models.generic import IDData
from vbl_aquarium.models.urchin import ParticleSystemModel

## Particle system
counter = 0
class ParticleSystem:
	"""Particle system
	
	Minimize the number of particle systems you create
	
	You cannot edit the number of particles in a system after creation
	
	Create separate particle systems when you need to use different materials
	"""
	
	def __init__(self, n, material = 'circle', positions = None, sizes = None, colors = None):
		"""Initialize particle system

		Parameters
		----------
		n : int
				Number of particles
		"""
		global counter
		counter += 1

		self.data = ParticleSystemModel(
			id= f'psystem{counter}',
			n = n,
			material= material,
			positions = [Vector3()] * n if positions is None else [utils.formatted_vector3(pos) for pos in utils.sanitize_list(positions, n)],
			sizes = [0.1] * n if sizes is None else utils.sanitize_list(sizes, n),
			colors = [Color()] * n if sizes is None else utils.sanitize_list(colors, n)
		)

		self._update()
		self.in_unity = True

	def _update(self):
		"""Push data to Urchin renderer
		"""
		client.sio.emit('urchin-particles-update', self.data.to_string())

	def delete(self):
		"""Delete this particle system and all its particles
		"""
		client.sio.emit('urchin-particles-delete', IDData(id= self.data.id).to_string())
		self.in_unity = False

	def set_material(self, material):
		"""Set the material of a particle system

		Options are
	 	- 'gaussian'
		- 'circle' (default)
		- 'circle-lit'
		- 'square'
		- 'square-lit'
		- 'diamond'
		- 'diamond-lit'

		Parameters
		---------- 
		material : string
			new material for all particles

		Examples
		--------
		>>>psystem1.set_material('circle')
		"""
		if self.in_unity == False:
			raise Exception("Particle system was deleted")
		
		self.data.material = utils.sanitize_string(material)
		self._update()

	def set_positions(self, positions):
		"""Set the positions of particles relative to the reference coordinate
		
		Parameters
		---------- 
		position : list of three floats
			(ap, ml, dv) coordinates in um

		Examples
		--------
		>>>psystem1.set_positions([5200,5700,330]) # move all particles to Bregma in CCF
		"""
		if self.in_unity == False:
			raise Exception("Particle system was deleted")
		
		positions = utils.sanitize_list(positions, self.data.n)
		self.data.positions = [utils.formatted_vector3([pos[0]/1000, pos[1]/1000, pos[2]/1000]) for pos in positions]
		
		self._update()

	def _set_positions(self, positions):
		"""Efficient position setting, for real-time applications

		Make sure your positions are in mm, not um

		ID should match psystem.data.id
		
		Parameters
		----------
		positions : Vector3List
				AP/ML/DV positions in *mm*
		"""
		if self.in_unity == False:
			raise Exception("Particle system was deleted")
		
		client.sio.emit('urchin-particles-positions', positions.to_string())

	def set_sizes(self, sizes):
		"""Set the sizes of particles
		
		Parameters
		---------- 
		size : float

		Examples
		--------
		>>>psystem1.set_sizes([0.02]) # 20 um 
		"""
		if self.in_unity == False:
			raise Exception("Particle system was deleted")
		
		sizes = utils.sanitize_list(sizes, self.data.n)
		self.data.sizes = [utils.sanitize_float(size)/1000 for size in sizes]

		self._update()

	def _set_sizes(self, sizes):
		"""Efficent size setting, for real-time applications

		ID should match psystem.data.id

		Parameters
		----------
		sizes : FloatList
				Sizes of particles in *mm*
		"""
		if self.in_unity == False:
			raise Exception("Particle system was deleted")
		
		client.sio.emit('urchin-particles-sizes', sizes.to_string())
	
	def set_colors(self, colors):
		"""Set the colors of particles
		
		Parameters
		---------- 
		colors : list
			hex or [r,g,b] colors

		Examples
		--------
		>>>psystem1.set_color(['#FFFFFF'])
		"""
		if self.in_unity == False:
			raise Exception("Particle system was deleted")
		
		colors = utils.sanitize_list(colors, self.data.n)

		self.data.colors = [utils.formatted_color(color) for color in colors]
		self._update()
	
	def _set_colors(self, colors):
		"""Efficient color setting, for real-time applications

		ID should match psystem.data.id

		Parameters
		----------
		colors : ColorList
				Colors of particles
		"""
		if self.in_unity == False:
			raise Exception("Particle system was deleted")
		
		client.sio.emit('urchin-particles-colors', colors.to_string())

def clear():
	"""Clear all particle systems
	"""
	client.sio.emit('Clear', 'particles')


"""Primitive meshes"""

from . import client
import warnings
from . import utils

from vbl_aquarium.models.urchin import *
from vbl_aquarium.models.generic import *

callback = None

def _neuron_callback(callback_data):
  if callback is not None:
    callback(callback_data)

counter = 0

## Primitive Mesh Renderer
class Mesh:
  """Mesh Object in Unity
  """

  def __init__(self, position= [0.0,0.0,0.0], scale= [1,1,1], color=[1,1,1],
               material = 'default', interactive = False):
    """Create a mesh object

    Note: this function should be avoided, use `urchin.meshes.create()`

    Parameters
    ----------
    position : list, optional
        position in ap/ml/dv Atlas coordinates, by default [0.0,0.0,0.0]
    scale : list, optional
        neuron scale, by default [1,1,1]
    color : str, optional
        by default '#FFFFFF'
    material : str, optional
        by default 'default'
    """
    global counter

    self.data = MeshModel(
      id = str(counter),
      shape = 'sphere',
      position = utils.formatted_vector3([x/1000 for x in position]),
      color = utils.formatted_color(color),
      scale = utils.formatted_vector3(scale),
      material = material,
      interactive = interactive
    )

    counter += 1

    self._update()
    self.in_unity = True

  def _update(self):
    """Serialize and update the data in the Urchin Renderer
    """
    client.sio.emit('urchin-meshes-update', self.data.to_string())

  def delete(self):
    """Deletes meshes

    Parameters
    ----------
    references object being deleted
      
	  Examples
	  --------
	  >>> cube_obj.delete() 
    """

    data = IDData
    data.id = self.data.id

    client.sio.emit('urchin-meshes-delete', data.to_string())
    self.in_unity = False
  
  def set_position(self, position):
    """Set the position of a mesh object. Position shoul dbe in AP/ML/DV coordinates in um

    Parameters
    ----------
    position : list of three floats
      (ap, ml, dv) in um
        
    Examples
    --------
    >>> cube_obj.set_position([2,2,2])
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    
    position_sanitized = utils.sanitize_vector3([position[0]/1000, position[1]/1000, position[2]/1000])
    self.data.position = utils.formatted_vector3(position_sanitized)

    self._update()
  
  def set_scale(self, scale):
    """Set the scale of mesh renderer

    Parameters
    ----------
    scale : list of three floats
      new scale of mesh
        
    Examples
    --------
    >>> cube_obj.set_scale([3,2,3])
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    
    scale_sanitized = utils.sanitize_vector3(scale)
    self.data.scale = utils.formatted_vector3(scale_sanitized)

    self._update()
  
  def set_color(self, color):
    """Set the color of mesh renderer

    Parameters
    ----------
    mesh_color : string hex color
      new hex color of mesh
        
    Examples
    --------
    >>> cube_obj.set_color("#000000")
    
    """ 
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    
    # color_sanitized = utils.sanitize_color(color)
    self.data.color = utils.formatted_color(color)

    self._update()

  def set_material(self, material):
    """Set the material of mesh renderer

    Parameters
    ----------
    mesh_material : string
      name of new material
        
    Examples
    --------
    >>> cube_obj.set_material('opaque-lit')
    
    """
    if self.in_unity == False:
      raise Exception("Object does not exist in Unity, call create method first.")
    
    material_sanitized = utils.sanitize_material(material)
    self.data.material = material_sanitized

    self._update()


def clear():
  client.sio.emit('Clear', 'mesh')

def create(num_objects, position= [0.0,0.0,0.0], scale= [1,1,1], color=[1,1,1],
               material = 'default', interactive = False):
  """Create multiple meshes

  Parameters
  ----------
  num_objects : int
      number of mesh objects to be created
  position : list, optional
      default position nullspace (ap, ml, dv), by default [0.0,0.0,0.0]
  scale : list, optional
      default scale, by default [1,1,1]
  color : list, optional
      default color, by default [1,1,1]
  material : str, optional
      default material, by default 'default'
  interactive : bool, optional
      default interactive state, by default False

	Examples
	--------
	>>> meshes = urchin.meshes.create(2)
  """
  mesh_objects = []
  for i in range(num_objects):
    mesh_objects.append(Mesh(position=position, color=color, scale=scale,
                             material=material, interactive=interactive))
  return(mesh_objects)

def delete(meshes_list):
  """Deletes meshes

  Parameters
  ----------
  meshes_list : list of mesh objects
	  list of meshes being deleted
      
	Examples
	--------
	>>> urchin.meshes.delete(meshes)
  """
  meshes_list = utils.sanitize_list(meshes_list)

  data = IDList(
    ids = [x.data.id for x in meshes_list]
  )

  client.sio.emit('urchin-meshes-deletes', data.to_string())

def set_positions(meshes_list, positions_list):
  """Set the positions of mesh renderers

  Parameters
  ----------
  meshes_list : list of mesh objects
	  list of meshes being set
  positions_list : list of list of three floats
    vertex positions of each mesh in um
      
	Examples
	--------
	>>> urchin.primitives.set_positions(cubes,[[3,3,3],[2,2,2]])
  """
  meshes_list = utils.sanitize_list(meshes_list)
  positions_list = utils.sanitize_list(positions_list)

  data = IDListVector3List(
    ids = [x.data.id for x in meshes_list],
    values = [utils.formatted_vector3(utils.sanitize_vector3([x[0]/1000, x[1]/1000, x[2]/1000])) for x in positions_list]
  )

  client.sio.emit('urchin-meshes-positions', data.to_string())

def set_scales(meshes_list, scales_list):
  """Set scale of mesh renderers

  Parameters
  ----------
  meshes_list : list of mesh objects
	  list of meshes being scaled
  scales_list : list of list of three floats
    new scales of each mesh
      
	Examples
	--------
	>>> urchin.primitives.set_scales(cubes,[[3,3,3],[2,2,2]])
  """
  meshes_list = utils.sanitize_list(meshes_list)
  scales_list = utils.sanitize_list(scales_list)

  data = IDListVector3List(
    ids = [x.data.id for x in meshes_list],
    values = [utils.formatted_vector3(utils.sanitize_vector3(x)) for x in scales_list]
  )

  client.sio.emit('urchin-meshes-scales', data.to_string())

def set_colors(meshes_list, colors_list):
  """Sets colors of mesh renderers

  Parameters
  ----------
  meshes_list : list of mesh objects
	  list of meshes undergoing color change
  colors_list : list of string hex colors
      new hex colors for each mesh
      
	Examples
	--------
	>>> urchin.primitives.set_colors(cubes,["#000000","#000000"])
	
  """
  meshes_list = utils.sanitize_list(meshes_list)
  colors_list = utils.sanitize_list(colors_list)

  data = IDListColorList(
    ids = [x.data.id for x in meshes_list],
    values = [utils.formatted_color(utils.sanitize_vector3(x)) for x in colors_list]
  )

  client.sio.emit('urchin-meshes-colors', data.to_string())

def set_materials(meshes_list, materials_list):
  """Sets materials of mesh renderers

  Parameters
  ----------
  meshes_list : list of mesh objects
	  list of meshes undergoing material change
  materials_list : list of strings
      name of new materials
      
	Examples
	--------
	>>> urchin.primitives.set_material(cubes,['opaque-lit','opaque-lit'])
	
  """
  meshes_list = utils.sanitize_list(meshes_list)
  materials_list = utils.sanitize_list(materials_list)

  data = IDListStringList(
    ids = [x.data.id for x in meshes_list],
    values = [utils.sanitize_material(x) for x in materials_list]
  )
      
  client.sio.emit('urchin-meshes-materials', data.to_string()) 


"""Volumetric datasets (x*y*z matrix)"""

from . import client
from . import utils
import numpy as np
import zlib
import json
import csv

counter = 0

CHUNK_LIMIT = 1000000

click_list = []
verbose = False

def _volume_click(data):
	"""Internal callback function
	"""
	click_pos = json.loads(data)
	if verbose:
		print(click_pos)
	
	click_list.append(click_pos)

def clear_clicks():
	"""Clear the volumes click list
	"""
	global click_list
	click_list = []

def save_clicks(fpath):
	"""Save the current click list to a CSV file.

	Use `urchin.volumes.clear_clicks()` before starting your click sequence.

	Parameters
	----------
	fpath : string
		Relative filepath
	"""
	global click_list
	# Extract headers and data
	headers = ['ap', 'ml', 'dv']
	data = [[entry[header] for header in headers] for entry in click_list]
	
	with open(fpath, mode='w', newline='') as file:
		writer = csv.writer(file)
		
		# Write headers
		writer.writerow(headers)
		
		# Write data
		writer.writerows(data)

def clear():
    """Clear all custom meshes
    """
    client.sio.emit('Clear','volume')

class Volume:
	"""Volumetric dataset represented in a compressed format by using a colormap to translate
	uint8 x/y/z data into full RGB color.

	Volumes should be created in (AP, ML, DV)
	"""
	def __init__(self, volume_data, colormap = None):
		"""_summary_

		Parameters
		----------
		volume_data : _type_
			_description_
		colormap : _type_, optional
			_description_, by default None
		"""
		global counter
		self.id = f'volume{counter}'
		counter += 1

		volume_data[np.isnan(volume_data)] = 255

		flattened_data = volume_data.flatten().astype(np.uint8).tobytes()
		compressed_data = zlib.compress(flattened_data)

		self.n_compressed_bytes = len(compressed_data)
		self.visible = True
		if colormap is not None:
			self.colormap = colormap
		else:
			self.colormap = ['#000000'] * 255

		self.update()

		# send data packets
		# split data into chunks
		n_chunks = int(np.ceil(self.n_compressed_bytes / CHUNK_LIMIT))
		print(f'Data fits in {n_chunks} chunks of 1MB or less')
		offset = 0
		for chunk in range(n_chunks):
			# get the data
			chunk_size = np.min((self.n_compressed_bytes - offset, CHUNK_LIMIT))

			chunk_data = {}
			chunk_data['name'] = self.id
			chunk_data['offset'] = int(offset)
			chunk_data['compressedByteChunk'] = list(compressed_data[offset : offset + chunk_size])
			client.sio.emit('SetVolumeData', json.dumps(chunk_data))

			offset += chunk_size

	def update(self):
		data = {}
		data['name'] = self.id
		data['nCompressedBytes'] = self.n_compressed_bytes
		data['visible'] = self.visible
		data['colormap'] = self.colormap

		client.sio.emit('UpdateVolume', json.dumps(data))

	def delete(self):
		client.sio.emit('DeleteVolume', self.id)

def compress_volume(volume_data, n_colors=254):
	"""Compress a volume of float data into a uint8 volume by quantiles.

	NaN values are mapped to 255 (transparent) for Urchin.

	This is required for use with the urchin.volume.Volume object type.

	Parameters
	----------
	volume_data : float volume
		3D matrix of float data
	n_colors : int (optional)
		Default to 254, number of un-reserved colors. 255 must always be reserved for NaN / transparency

	Returns
	-------
	(uint8 volume, float[] map)
	"""
	quantiles = np.quantile(volume_data.flatten(), np.linspace(0,1,n_colors))

	out = np.digitize(volume_data, quantiles, right=True).astype(np.uint8)
	out[np.isnan(volume_data)] = 255

	return out.astype(np.uint8), quantiles

def colormap(colormap_name='greens', reserved_colors=[], datapoints=None):
	"""Build a colormap

	This function has two parts:
	1. It builds a standard colormap in indexes 0->n_colors
	2. It leaves "reserved" colors at the end, by default this is just 255 which becomes 
	transparent in Urchin. But you can add a list of additional colors which will be added
	to the end of the colormap. The order will match the list you pass in, so e.g.
	3. If you pass in datapoints, it will generate a non-uniform colormap going from the min
	to maximum value.

	indexes: 	[0->252, 	253->254, 				255]
	colors: 	[greens, 	your reserved colors, 	transparent]

	Colormap options
	----------
		reds: 0->255 R channel
		greens: 0->255 G channel
		blues: 0->255 B channel

	Parameters
	----------
	colormap_name : str, optional
		_description_, by default 'greens'
	reserved_colors : _type_, optional
		_description_, by default None

	Returns
	-------
	list of string
		List of colormap hex colors in Urchin-compatible format
	"""
	colormap = []

	n_unreserved = 254 - len(reserved_colors)

	datapoints = (datapoints - np.min(datapoints)) / (np.max(datapoints) - np.min(datapoints))
	datapoints = datapoints / np.max(datapoints)

	for i in range(0, n_unreserved):
		if datapoints is not None:
			v = int(np.round(datapoints[i] * 255))
		else:
			v = int(np.round(i / n_unreserved * 255))

		if colormap_name == 'reds':
			colormap.append(utils.rgba_to_hex((v,0,0,255)))
		elif colormap_name == 'greens':
			colormap.append(utils.rgba_to_hex((0,v,0,255)))
		elif colormap_name == 'blues':
			colormap.append(utils.rgba_to_hex((0,0,v,255)))
		else:
			raise Exception(f'{colormap_name} is not a valid colormap option')

	colormap.extend(reserved_colors)

	return colormap


"""Lines"""

from . import client
import warnings
from . import utils

from vbl_aquarium.models.urchin import LineModel
from vbl_aquarium.models.generic import IDData

counter = 0

def clear():
    """Clear all custom meshes
    """
    client.sio.emit('Clear','lines')

class Line:
  def __init__(self, positions= [[0.0,0.0,0.0]], color= [1, 1, 1]):
    global counter
    counter += 1

    self.data = LineModel(
      id = f'l{counter}',
      positions = positions,
      color=utils.formatted_color(color)
    )

    self._update()
    self.in_unity = True

  def _update(self):
    client.sio.emit('urchin-line-update', self.data.to_string())

  def delete(self):
    """Deletes lines

    Examples
    >>>l1.delete()
    """
    client.sio.emit('urchin-line-delete', IDData(id=self.data.id).to_string())
    self.in_unity = False

  def set_positions(self, positions):
    """Set the positions of line renderer
    
    Parameters
    ---------- 
    position : list of vector3 [[ap,ml,dv],[ap,ml,dv]]
        vertex positions of the line in the ReferenceAtlas space (um)

    Examples
    --------
    >>>l1.set_position([[0, 0, 0],[13200,11400,8000]])
    """
    if self.in_unity == False:
      raise Exception("Line does not exist in Unity, call create method first.")
    
    for i, vec3 in enumerate(positions):
      positions[i] = utils.formatted_vector3(utils.sanitize_vector3(vec3))
    self.data.positions = positions

    self._update()

  def set_color(self, color):
    """Set the color of line renderer
    
    Parameters
    ---------- 
    color : string hex color
        new color of the line

    Examples
    --------
    >>>l1.set_color('#000000')
    """
    if self.in_unity == False:
      raise Exception("Line does not exist in Unity, call create method first.")
    
    self.data.color = utils.formatted_color(utils.sanitize_color(color))

    self._update()

def create (n):
  """Create Line objects

  Parameters
  ----------
  n : int
      Number of objects to create
  """
  lines_list = []
  
  for i in range(n):
    line = Line()
    lines_list.append(line)

  return lines_list

def delete (lines_list):
  """Deletes lines
  
  Parameters
  ---------- 
  lines_list : list of Line objects
      list of lines to be deleted

  Examples
  --------
  >>> lines.delete()
  """
  
  for line in lines_list:
    line.delete()




from . import client
import requests
import hashlib

from vbl_aquarium.models.dock import BucketRequest, SaveRequest, LoadRequest, DockModel

# Define the API endpoint URL
api_url = "http://localhost:5000"

active_bucket = None
password_hash = None
test_token = "c503675a-506c-48c0-9b5e-5265e8260a06"

callback_filename = None

def _save_callback(data):
    global callback_filename
    print(f"(dock) Save callback received, saving data to {callback_filename}")
    # data will be a serialized LoadModel, but we can just ignore that and save it
    with open(callback_filename, 'rb') as file:
        file.write(data)

def create_bucket(bucket_name, password, api_url = api_url, token = test_token):
    """Create a new bucket for storing data

    Parameters
    ----------
    bucket_name : str
        Folder data will be stored in
    password : str
        Passwords are hashed client-side
    api_url : str, optional
        host:port, by default api_url

    Returns
    -------
    str
        bucket name
    """
    global active_bucket
    global password_hash

    headers = {
        "Content-Type": "application/json"
    }

    # Set the API URL in Urchin
    api_data = DockModel(
        dock_url=api_url
    )

    client.sio.emit('urchin-dock-data', api_data.to_string())

    # Request new bucket
    create_url = f'{api_url}/create/{bucket_name}'

    active_bucket = bucket_name
    password_hash = hash256(password)

    data = BucketRequest(
        token = test_token,
        password = password_hash
    )

    print(f'Attempting to create {create_url}')
    response = requests.post(create_url, data=data.model_dump_json(), headers=headers)

    # Check the response
    if response.status_code == 201:
        print(response.text)
    else:
        print("Error:", response.status_code, response.text)

    return bucket_name


def save(filename = None, bucket_name = None, password = None):
    """Save all current data, either to a file or to a cloud bucket.

    Either the filename or bucket/password are required

    Parameters
    ----------
    bucket_name : str
    password : str
    """
    global active_bucket
    global password_hash

    if filename is not None:
        global callback_filename
        callback_filename = filename
    else:
        check_and_store(bucket_name, password)

    data = SaveRequest(
        filename= "" if filename is None else filename,
        bucket= "" if active_bucket is None else active_bucket,
        password= "" if password_hash is None else password_hash
    )

    client.sio.emit('urchin-save', data.to_string())

def load(filename = None, bucket_name = None, password= None):
    """Load all data from a bucket

    Either a filename or bucket/password are required

    Parameters
    ----------
    filename : str, optional
    bucket_name : str, optional
    password : str, optional
    """
    global active_bucket
    global password_hash

    if filename is not None:
        raise Exception("Not implemented yet")
    
    check_and_store(bucket_name, password)

    data = LoadRequest(
        filename= "" if filename is None else filename,
        bucket= "" if active_bucket is None else active_bucket,
        password= "" if password_hash is None else password_hash
    )

    client.sio.emit('urchin-load', data.to_string())

def check_and_store(bucket_name, password):
    global active_bucket
    global password_hash
    bucket_name = bucket_name if bucket_name is not None else active_bucket
    if bucket_name is None:
        raise Exception("Bucket name is required if there isn't one already stored.")
    elif active_bucket is None:
        active_bucket = bucket_name

    password = hash256(password) if password is not None else password_hash
    if password is None:
        raise Exception("Password is required if there isn't one already stored.")
    elif password_hash is None:
        password_hash = password

def hash256(password):
    return hashlib.sha256(password.encode()).hexdigest()



"""Custom 3D Objects"""

from . import client
from . import utils
import json

from vbl_aquarium.models.urchin import CustomMeshModel
from vbl_aquarium.models.generic import IDData

counter = 0

class CustomMesh:
    """Custom 3D object
    """
    
    def __init__(self, vertices, triangles, normals = None):
        """Create a Custom 3D object based on a set of vertices and triangle faces

        Unity can automatically calculate the normals to create a convex object, or you can pass them yourself.

        Parameters
        ----------
        vertices : list of vector3
            Vertex coordinates, the x/y/z directions will correspond to AP/DV/ML if your object was exported from Blender
        triangles : list of int
            Triangle vertex 
        normals : list of vector3, optional
            Normal directions, by default None
        """
        global counter
        counter += 1

        self.data = CustomMeshModel(
            id = str(counter),
            vertices = [utils.formatted_vector3(x) for x in vertices],
            triangles = [utils.formatted_vector3(x) for x in triangles],
            normals = normals
        )

        self._update()
        self.in_unity = True

    def _update(self):
        client.sio.emit('urchin-custommesh-update', self.data.to_string())

    def delete(self):
        """Destroy this object in the renderer scene
        """

        data = IDData(
            id = self.data.id
        )

        client.sio.emit('urchin-custommesh-delete', data.to_string())
        self.in_unity = False

    def set_position(self, position = [0,0,0], use_reference = True):
        """Set the position relative to the reference coordinate

        Note that the null transform is active by default in Urchin, therefore directions are the CCF defaults:
        AP+ posterior, ML+ right, DV+ ventral

        By default objects are placed with their origin at the reference (Bregma), disabling this
        places objects relative to the Atlas origin, which is the (0,0,0) coordinate in the top, front, left
        corner of the atlas space.

        Parameters
        ----------
        position : vector3
            AP/ML/DV coordinate relative to the reference (defaults to [0,0,0] when unset)
        use_reference : bool, optional
            whether to use the reference coordinate, by default True
        """

        self.data.position = utils.formatted_vector3(position)
        self.data.use_reference = use_reference

        self._update()

    def set_scale(self, scale = [1, 1, 1]):
        """_summary_

        Parameters
        ----------
        scale : list, optional
            _description_, by default [1, 1, 1]
        """

        self.data.scale = utils.formatted_vector3(scale)

        self._update()

def clear():
    """Clear all custom meshes
    """
    client.sio.emit('Clear','custom')